#!/usr/bin/env python3
import os
import re
import sys
import time
import array
import getopt
import subprocess
import syslog
from syslog import syslog as log

syslog.setlogmask(syslog.LOG_UPTO(syslog.LOG_INFO))
syslog.openlog(os.path.basename(__file__), syslog.LOG_PID | syslog.LOG_NDELAY | syslog.LOG_PERROR, syslog.LOG_LOCAL1)

VERSION = "1.1.51"


class Stm32CRC:
    def __init__(self):
        pass

    @staticmethod
    def process_word(data, crc=0xffffffff):
        crc_poly = 0x04C11DB7
        if len(data) < 4:
            d_array = array.array('B', data)
            for x in range(0, 4 - len(data)):
                d_array.insert(0, 0)
            d_array.reverse()
            data = d_array.tostring()

        d = array.array('I', data)[0]
        crc = crc ^ d

        for i in range(0, 32):
            if (crc & 0x80000000) != 0:
                crc = (crc << 1) ^ crc_poly
            else:
                crc = (crc << 1)

        result = crc & 0xffffffff
        return result

    def process_buffer(self, buf, c=0xffffffff):
        word_count = int(len(buf) / 4)
        if len(buf) % 4 != 0:
            word_count += 1

        crc = c
        for i in range(0, word_count):
            crc = self.process_word(buf[i * 4: (i + 1) * 4], crc)
        return crc

    def crc32(self, filename):
        return self.process_buffer(open(filename, 'rb').read())


class PqmStmFlash:
    def __init__(self, cmd_path, fwfile, debug=False, force=False, reset=False):
        self.debug = debug
        self.forceflash = force
        self.fwfile = os.path.realpath(fwfile)
        self.settingsfile = "{}/pqmstm32flash.crc".format(os.path.dirname(os.path.realpath(__file__)))
        self.smt32flash_cmd_template = "%s -R -i '166&90,,,,,-90,,,,:-166&90,,,-90' -a 0x48 %s %s /dev/i2c-0"
        self.smt32flash_cmd_template_reset = "%s -R -i '166&90,-90,:-166&90,-90' -a 0x48 /dev/i2c-0"
        self.stm32flash_path = cmd_path
        self.stm32flash_wrt = self.smt32flash_cmd_template % (self.stm32flash_path, "-v -w", self.fwfile)
        self.stm32flash_rst = self.smt32flash_cmd_template_reset % (self.stm32flash_path)
        if self.fwfile != "":
            self.fwfilesize = os.path.getsize(self.fwfile)
            self.stm32flash_get_crc = \
                self.smt32flash_cmd_template % (self.stm32flash_path, "-C -S", "0x08000000:%s" % self.fwfilesize)

            self.oldfwcrc = 0
            self.oldfwfilesize = 0
            crc = Stm32CRC()
            self.fwfilecrc = crc.crc32(self.fwfile)
            self.need_flash = True

            self.do_settings()

            if self.forceflash:
                self.need_flash = True

    def printd(self, data):
        if self.debug:
            print(data)

    def do_settings(self):
        if os.path.exists(self.settingsfile):
            crcfile = open(self.settingsfile, "r+")
            ofwcrc = crcfile.readline()
            self.oldfwcrc = int(ofwcrc, 16)
        else:
            crcfile = open(self.settingsfile, "w")
            self.oldfwcrc = self.get_crc_stm32()
            if self.oldfwcrc != 0xff:
                crcfile.write(hex(self.oldfwcrc))
                crcfile.write("\n")

        if self.oldfwcrc == self.fwfilecrc:
            self.need_flash = False
        else:
            crcfile.seek(0)
            crcfile.write(hex(self.fwfilecrc))
            crcfile.write("\n")

        crcfile.close()

    def get_crc_stm32(self):
        retval, retstring = subprocess.getstatusoutput(self.stm32flash_get_crc)
        if retval == 0:
            crc_list = \
                re.findall(r'CRC\((0[xX][0-9a-fA-F]+).(0[xX][0-9a-fA-F]+).*(0[xX][0-9a-fA-F]+)', retstring)[0]
            retval = int(crc_list[2], 16)
        return retval

    def do_main(self):
        retval = 2
        if self.need_flash:
            self.printd("flash using cmd: " + self.stm32flash_wrt)
            retval, retstring = subprocess.getstatusoutput(self.stm32flash_wrt)
            self.printd(retstring)
            self.printd("Exit code: {}".format(retval))
        return retval

    def do_reset(self):
        retval = 2

        self.printd("reset using cmd: " + self.stm32flash_rst)
        retval, retstring = subprocess.getstatusoutput(self.stm32flash_rst)
        self.printd(retstring)
        self.printd("Exit code: {}".format(retval))
        return retval

    def delete_crc_file(self):
        os.remove(self.settingsfile)


def usage(smt32path):
    print("%s \n"
          "   -f stm32firmware.bin  # firmwarefile\n"
          "  [-s stm32flash binary] # stm32flash path default %s\n"
          "  [-d]                   # debug prints\n"
          "  [-i]                   # ignore crc and force flash\n"
          "   -r                    # only reset STM32" % (sys.argv[0], smt32path))
    sys.exit(-1)


def main(argv):
    fwfile = ""
    smt32flashpath = "/usr/bin/stm32flash"
    debug_prints = False
    force_flash = False
    reset_stm = False

    try:
        opts, args = getopt.getopt(argv, "hf:s:dir", ["help", "fwfile=", "smt32flashpath=", "debug", "ignore", "reset"])
    except getopt.GetoptError:
        usage(smt32flashpath)

    for opt, arg in opts:
        if opt in ("-h", "--help"):
            usage(smt32flashpath)
            sys.exit()
        elif opt == '-d':
            debug_prints = True
        elif opt in ("-f", "--fwfile"):
            fwfile = arg
        elif opt in ("-s", "--smt32flashpath"):
            smt32flashpath = arg
        elif opt in ("-i", "--ignore"):
            force_flash = True
        elif opt in ("-r", "--reset"):
            reset_stm = True

    if not reset_stm:
        if fwfile == "":
            log(syslog.LOG_CRIT, "no firmware file given")
            usage(smt32flashpath)

        if not os.path.exists(fwfile):
            log(syslog.LOG_CRIT, "no firmware file: {}".format(fwfile))
            sys.exit(-1)

        if not os.path.exists(smt32flashpath):
            log(syslog.LOG_CRIT, "stm32flash not found in folder {}".format(smt32flashpath))
            sys.exit(-1)

    pqmstm = PqmStmFlash(smt32flashpath, fwfile, debug=debug_prints, force=force_flash, reset=reset_stm)

    if reset_stm:
        retval = pqmstm.do_reset()
    else:
        retval = 4
        for i in range(0, 3):
            retval = pqmstm.do_main()
            if retval == 2 or retval == 0:
                break
            time.sleep(2)

        fwcrc = [hex(pqmstm.oldfwcrc), hex(pqmstm.fwfilecrc)]
        if retval == 0:
            log(syslog.LOG_INFO, "flash successfull ({} -> {})".format(*fwcrc))
        elif retval == 2:
            log(syslog.LOG_INFO, "flash not needed ({} -> {})".format(*fwcrc))
        else:
            pqmstm.delete_crc_file()
            log(syslog.LOG_CRIT, "flash not successfull ({} -> {})".format(*fwcrc))

    sys.exit(retval)


if __name__ == '__main__':
    log(syslog.LOG_INFO, "pqmstm32flash version {}".format(VERSION))
    main(sys.argv[1:])
